"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.System = void 0;
require("reflect-metadata");
const toposort_1 = require("./toposort");
const dependenciesKey = Symbol('component:dependencies');
const constructorKey = Symbol('components:constructor');
const systemComponents = Symbol('components:systemComponents');
function ComponentUsing(dependencies = [], build) {
    return (target, key) => {
        const meta = Reflect.getMetadata(systemComponents, target) || [];
        Reflect.defineMetadata(systemComponents, [...meta, key], target);
        Reflect.defineMetadata(dependenciesKey, dependencies, target, key);
        Reflect.defineMetadata(constructorKey, build, target, key);
    };
}
function getSystemDecenciesGraph(system) {
    const components = Reflect.getMetadata(systemComponents, system);
    return components.map(key => ({ id: key, edges: system._getDependencies(key) }));
}
function getInitializationOrder(system) {
    const dependencyGraph = getSystemDecenciesGraph(system);
    return toposort_1.topoSort(dependencyGraph);
}
function startSystem(system) {
    return __awaiter(this, void 0, void 0, function* () {
        const initializeOrder = getInitializationOrder(system);
        for (const componentName of initializeOrder) {
            const componentBuilder = system._getConstructor(componentName);
            const currentComponent = componentBuilder(system);
            try {
                system._setComponent(componentName, currentComponent);
                if (currentComponent.start !== undefined) {
                    system.logger.info(`Starting ${componentName}...`);
                    yield currentComponent.start();
                    system.logger.info(`${componentName} started with success!`);
                }
            }
            catch (e) {
                system.logger.error(`Error starting ${componentName}`);
                system.logger.error(e);
                throw e;
            }
        }
        return system;
    });
}
function stopSystem(system) {
    return __awaiter(this, void 0, void 0, function* () {
        const stopOrder = getInitializationOrder(system).reverse();
        system.logger.info(`Stopping System: Order ${stopOrder}`);
        for (const componentName of stopOrder) {
            const component = system._getComponent(componentName);
            system.logger.info(`Stopping ${componentName}...`);
            try {
                if (component !== undefined && component !== null)
                    yield component.stop();
            }
            catch (e) {
                system.logger.error(`Error stopping component ${componentName}`, e);
                console.error(e);
            }
            system.logger.info(`${componentName} stopped!`);
        }
        return system;
    });
}
class System {
    constructor() {
        /// You can provide your custom logger here
        this.logger = {
            info: (message) => console.log(message),
            error: (message, e) => console.error(message, e),
        };
    }
    static Using(dependencies, build) {
        return ComponentUsing(dependencies, build);
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            return startSystem(this);
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            return stopSystem(this);
        });
    }
    // Do not use directly anything down below
    _getComponent(componentName) {
        return this[componentName];
    }
    _getDependencies(componentName) {
        return Reflect.getMetadata(dependenciesKey, this, componentName) || [];
    }
    _getConstructor(componentName) {
        return Reflect.getMetadata(constructorKey, this, componentName);
    }
    _setComponent(componentName, comp) {
        this[componentName] = comp;
    }
}
exports.System = System;
//# sourceMappingURL=system.js.map